#include "Game.h"
#include "Trap.h"
#include "Boss.h"
const int Game::TRANSITION_TIME = 30;
const int Game::GAME_OVER_FADE_TIME = 120;
const int Game::ROOM_SIZE = 288;
const int Game::DESCENT_TIME = 120;

Game::Game(sf::RenderWindow& window):
    window(window),
    onBuffer(true),
    player(nullptr),
    currLevel(nullptr),
    depth(0),
    transitioning(false),
    currRoom(nullptr),
    nextRoom(nullptr),
    showBoundingBoxes(false),
    fullScreen(false),
    menuPlayerSprite(ResourceManager::getTexture("player_walk.png"), 32, 32, -1, 3, 0, 1, true),
    menuPlayerIdleSprite(ResourceManager::getTexture("player_sleep.png"), 32, 32, -1, 30, 0, 0, true)
{
    sf::View view = window.getDefaultView();
    view.setSize(400, 288);
    view.setCenter(200, 144);
    buffer.create(288, 288);

    window.setView(view);

    input.loadDefaults();

    resetGame();
}

Game::~Game()
{
    if(currLevel)
    {
        delete currLevel;
    }
}

void Game::resetGame()
{
    frame = 0;
    if(currLevel)
    {
        delete currLevel;
        currLevel = nullptr;
    }
    depth = 0;

    onMainMenu = true;
    paused = false;
    startingGame = false;
    descentTimer = DESCENT_TIME;
    menuShakeAmount = 100.f;
    menuPlayerLocation = 200 - menuPlayerSprite.getSize().x / 2;
}
void Game::step()
{
    handleInput();
    if(onMainMenu)
    {
        mainMenuStep();
        return;
    }
    frame++;
    if(transitioning)
    {

    }
    else
    {
        currRoom->step();
        if(player->die)
        {
            if(gameOverFadeTimer > 0)
            {
                gameOverFadeTimer--;
            }
            else
            {
                if(input.isPressed(Input::ACTION_ATTACK))
                {
                    resetGame();
                }
                player->x += (200 - player->x - player->w/2) / 30.f;
                player->y += (154 - player->y - player->h/2) / 30.f;
            }
        }
    }
    handleTransition();
    if(sf::Keyboard::isKeyPressed(sf::Keyboard::F1))
    {
        showBoundingBoxes = true;
    }
}


void Game::draw()
{
    window.clear(sf::Color::Black);
    if(onMainMenu)
    {
        drawMainMenu();
        window.display();
        return;
    }
    if(transitioning)
    {
        buffer.clear(sf::Color::Black);
        currRoom->draw(buffer, showBoundingBoxes);
        buffer.display();
        sf::Sprite sprite(buffer.getTexture());

        // easing function~
        float amount = -1 /2.f * (cos(PI*transitionTimer/(float)TRANSITION_TIME) - 1);

        sprite.setPosition(amount * -transitionDir.x * ROOM_SIZE,
                           amount * -transitionDir.y * ROOM_SIZE);
        window.draw(sprite);


        buffer.clear(sf::Color::Black);
        nextRoom->draw(buffer, showBoundingBoxes);
        buffer.display();
        sf::Sprite spriteNext(buffer.getTexture());
        spriteNext.setPosition(amount * -transitionDir.x * ROOM_SIZE + ROOM_SIZE * transitionDir.x,
                            amount * -transitionDir.y * ROOM_SIZE + ROOM_SIZE * transitionDir.y);
        window.draw(spriteNext);
    }
    else
    {
        buffer.clear(sf::Color::Black);
        currRoom->draw(buffer, showBoundingBoxes);
        buffer.display();
        sf::Sprite sprite(buffer.getTexture());
        sprite.setPosition(0, 0);
        window.draw(sprite);
    }
    drawGui();
    window.display();
}

void Game::startGame()
{
    ResourceManager::playMusic("bgm.ogg");
    player = new Player(ROOM_SIZE/2-Player::WIDTH/2, ROOM_SIZE/2-Player::HEIGHT/2);

    currLevel = new Level(0);
    currRoom = currLevel->getStartRoom();
    currRoom->addEntity(player);

    currRoom->explored = true;
    onMainMenu = false;
    gameOverFadeTimer = GAME_OVER_FADE_TIME;
}

void Game::mainMenuStep()
{
    if(input.isPressed(Input::ACTION_ATTACK))
    {
        startingGame = true;
    }
    if(startingGame && !input.binding)
    {
        if(menuPlayerLocation < 400)
        {
            menuPlayerLocation += 3.3;
        }
        else
        {
            startGame();
        }
    }
}

void Game::drawMainMenu()
{
    if(input.binding)
    {
        stringstream prompt;
        if(!input.controllerChosen)
        {
            prompt << "Hold any key or a gamepad button to choose Keyboard or Gamepad";
        }
        else
        {
            prompt << "Hold a button for " << input.actionNames[input.currentlyBinding];
        }
        sf::Text promptText(prompt.str(), ResourceManager::getDefaultFont(), 8);
        promptText.setPosition(10, 16);
        promptText.setColor(sf::Color::White);
        window.draw(promptText);
        float progress = (input.controllerChosen ? input.bindTimer : input.controllerChooseTimer) / (float)Input::BIND_TIME;
        sf::RectangleShape bar({400 * progress, 10});
        progress = progress * progress;
        bar.setPosition(0, 0);
        bar.setFillColor(sf::Color::White);
        window.draw(bar);
    }
    if(!paused)
    {
        if(startingGame)
        {
            menuPlayerSprite.step();
            menuPlayerSprite.setPosition(menuPlayerLocation, 145);
            menuPlayerSprite.draw(window);
        }
        else
        {
            menuPlayerIdleSprite.step();
            menuPlayerIdleSprite.setPosition(menuPlayerLocation, 145);
            menuPlayerIdleSprite.draw(window);
        }
        if(!input.binding)
        {
            stringstream ss;
            ss << "Press " << input.getBindName(Input::ACTION_ATTACK) <<" to start!";
            string text = ss.str();
            for(unsigned i = 0; i < text.size(); i++)
            {
                sf::Text charText(text[i], ResourceManager::getDefaultFont(), 8);
                charText.setPosition(200 - (text.size()/2.f * 10) + i * 10 + (rand()/(float)RAND_MAX - 0.5f) * menuShakeAmount,
                                    270 + (rand()/(float)RAND_MAX - 0.5f) * menuShakeAmount);
                charText.setColor(sf::Color::White);
                window.draw(charText);
            }
            ss.str("");
            ss << "Press F1 to rebind controls!";
            text = ss.str();
            for(unsigned i = 0; i < text.size(); i++)
            {
                sf::Text charText(text[i], ResourceManager::getDefaultFont(), 8);
                charText.setPosition(200 - (text.size()/2.f * 10) + i * 10 + (rand()/(float)RAND_MAX - 0.5f) * menuShakeAmount,
                                    7 + (rand()/(float)RAND_MAX - 0.5f) * menuShakeAmount);
                charText.setColor(sf::Color::White);
                window.draw(charText);
            }
        }

    }
    if(startingGame)
    {
        menuShakeAmount *= 1.4;
    }
    else if(menuShakeAmount > 1)
    {
        menuShakeAmount *= 0.9;
    }
}

void Game::drawGui()
{
    sf::RectangleShape bgBox({400 - ROOM_SIZE, ROOM_SIZE});
    bgBox.setPosition(ROOM_SIZE, 0);
    bgBox.setFillColor(sf::Color(0, 0, 0));
    window.draw(bgBox);
    bgBox.setSize({1, ROOM_SIZE});
    bgBox.setFillColor(sf::Color::White);
    window.draw(bgBox);

    currLevel->drawMinimap(window, ROOM_SIZE + 18, 22);
    player->drawStats(window, ROOM_SIZE + 26, 140);

    sf::Text text;
    stringstream leveltext;
    leveltext << "Level " << (depth + 1);
    text.setString(leveltext.str());
    text.setPosition(ROOM_SIZE + 18, 10);
    text.setFont(ResourceManager::getDefaultFont());
    text.setCharacterSize(8);
    window.draw(text);

    if(player->die)
    {
        sf::RectangleShape fade({(float)window.getSize().x, (float)window.getSize().y});
        float fadeAmount = 1 - gameOverFadeTimer / (float)GAME_OVER_FADE_TIME;
        fade.setFillColor(sf::Color(0, 0, 0, 255 * fadeAmount));
        window.draw(fade);
        player->draw(window);

        if(gameOverFadeTimer == 0)
        {
            string gameOver = "GAME OVER";
            for(unsigned i = 0; i < gameOver.size(); i++)
            {
                sf::Text gameOverText(gameOver[i], ResourceManager::getDefaultFont(), 16);
                gameOverText.setPosition(110 + i * 20, 100 + sinf((frame + i * 20.f) / 60.f) * 7);
                gameOverText.setColor(sf::Color::White);
                window.draw(gameOverText);
            }
            sf::Text collectedText("YOU DIED WITH", ResourceManager::getDefaultFont(), 8);
            collectedText.setColor(sf::Color::White);
            collectedText.setPosition(200 - collectedText.getLocalBounds().width/2, 236);
            window.draw(collectedText);
            player->drawItemWave(window, 206, 263);
        }
    }
}

void Game::handleInput()
{
    input.step();
    if(!onMainMenu && player && !player->descending)
    {
        if(input.isHeld(Input::ACTION_LEFT))
        {
            player->moveLeft();
        }
        if(input.isHeld(Input::ACTION_UP))
        {
            player->moveUp();
        }
        if(input.isHeld(Input::ACTION_DOWN))
        {
            player->moveDown();
        }
        if(input.isHeld(Input::ACTION_RIGHT))
        {
            player->moveRight();
        }
        if(input.isPressed(Input::ACTION_ATTACK))
        {
            player->attack();
        }
        if(input.isPressed(Input::ACTION_RUN))
        {
            player->startRun();
        }
        if(input.isReleased(Input::ACTION_RUN))
        {
            player->stopRun();
        }
        sf::Vector2i mPos = sf::Mouse::getPosition(window);

        if(sf::Keyboard::isKeyPressed(sf::Keyboard::F12))
        {
            resetGame();
        }
    }
    else
    {
        static bool prevF1Pressed = false;
        if(sf::Keyboard::isKeyPressed(sf::Keyboard::F1) && !prevF1Pressed)
        {
            input.reset();
        }
        prevF1Pressed = sf::Keyboard::isKeyPressed(sf::Keyboard::F1);
    }
    static bool prevPressed = false;
    if(sf::Keyboard::isKeyPressed(sf::Keyboard::F10) && !prevPressed)
        {
            fullScreen = !fullScreen;
            if(fullScreen)
            {
                vector<sf::VideoMode> modes = sf::VideoMode::getFullscreenModes();
                sf::VideoMode mode = modes[0];
                for(unsigned i = 0; i < modes.size(); i++)
                {
                    if(modes[i].width == 800 && modes[i].height == 600 && modes[i].bitsPerPixel == 32)
                    {
                        mode = modes[i];
                    }
                }
                window.create(mode, "IRONE", sf::Style::Fullscreen);
                sf::View view = window.getDefaultView();
                view.setSize(400, 288);
                view.setCenter(200, 144);
                window.setView(view);
            }
            else
            {
                window.create(sf::VideoMode(800, 576), "IRONE");
                sf::View view = window.getDefaultView();
                view.setSize(400, 288);
                view.setCenter(200, 144);
                window.setView(view);
            }
        }

        prevPressed = sf::Keyboard::isKeyPressed(sf::Keyboard::F10);
}

void Game::handleTransition()
{
    if(!transitioning)
    {
        if(player->transition)
        {
            transitionDir = player->transitionDir;
            sf::Vector2i destinationCoords = transitionDir + sf::Vector2i(currRoom->x, currRoom->y);
            nextRoom = currLevel->getRoom(destinationCoords.x, destinationCoords.y);
            assert(nextRoom);

            transitioning = true;
            transitionTimer = 0;
            currRoom->removeEntity(player);
            nextRoom->addEntity(player);
            player->x -= transitionDir.x * ROOM_SIZE;
            player->y -= transitionDir.y * ROOM_SIZE;
            nextRoom->explored = true;
        }
        else if (player->descending)
        {
            int halfway = DESCENT_TIME / 2;
            descentTimer--;
            if(descentTimer > halfway)
            {
                float amt = 1 - (descentTimer - halfway) / (float)halfway;
                player->descentPercent = amt;
                player->leaving = true;
                player->x += (144 - player->x - player->w/2) / 10.f;
                player->y += (144 - player->y - player->h/2) / 10.f;
            }
            else if(descentTimer == halfway)
            {
                depth++;
                currRoom->removeEntity(player);
                if (currLevel) //don;t need this if but let's be safe
                {
                    delete currLevel;
                }
                currLevel = new Level(depth);
                currRoom = currLevel->getStartRoom();
                currRoom->addEntity(player);
                currRoom->explored = true;
            }
            else if(descentTimer > 0)
            {
                float amt = 1.f - (descentTimer / (float)halfway);
                player->descentPercent = amt;
                player->leaving = false;
            }
            else
            {
                player->descending = false;
                descentTimer = DESCENT_TIME;
            }
        }
    }
    else
    {
        if(transitionTimer < TRANSITION_TIME)
        {
            transitionTimer++;
        }
        else
        {
            transitioning = false;
            currRoom = nextRoom;
            nextRoom = nullptr;
        }
    }
}
