#include "Level.h"
#include "Room.h"

const int Level::WIDTH = 5;
const int Level::HEIGHT = 5;

Level::Level(int floor):
    w(WIDTH),
    h(HEIGHT),
    rooms(nullptr),
    startRoom(nullptr)
{
    rooms = new Room**[w];
    for(int x = 0; x < w; x++)
    {
        rooms[x] = new Room*[h];
        for(int y = 0; y < h; y++)
        {
            rooms[x][y] = nullptr;
        }
    }

    generate(floor);
}

Level::~Level()
{
    //dtor
    for(int x = 0; x < w; x++)
    {
        for(int y = 0; y < h; y++)
        {
            if(rooms[x][y])
            {
                delete rooms[x][y];
                rooms[x][y] = nullptr;
            }
        }
        delete[] rooms[x];
    }
    delete[] rooms;
}


void Level::generate(int floor)
{
    static int numRooms = 11; //9 plus start and boss

    sf::Vector2i startPos(rand()%w, rand()%h);
    startRoom = new Room(startPos.x, startPos.y, floor);
    rooms[startPos.x][startPos.y] = startRoom;
    vector<sf::Vector2i> roomPositions;
    roomPositions.push_back(startPos);

    int roomCount = 1; //include start
    while(roomCount < numRooms)
    {
        sf::Vector2i parentPos = roomPositions[rand()%roomPositions.size()];
        sf::Vector2i targetPos = parentPos;
        int dir = rand()%4;
        switch(dir)
        {
        case 0:
            targetPos.x--;
            break;
        case 1:
            targetPos.x++;
            break;
        case 2:
            targetPos.y--;
            break;
        case 3:
            targetPos.y++;
            break;
        }
        if(targetPos.x < 0 || targetPos.y < 0 ||
           targetPos.x >= w || targetPos.y >= h ||
           rooms[targetPos.x][targetPos.y] != nullptr)
        {
            continue;
        }
        Room* parent = rooms[parentPos.x][parentPos.y];
        rooms[targetPos.x][targetPos.y] = new Room(targetPos.x, targetPos.y, floor, parent);
        linkRooms(parentPos, targetPos);
        roomPositions.push_back(targetPos);
        roomCount++;
    }
    //place the boss room.
    sf::Vector2i bossPos = roomPositions[roomCount-1];
    //we do the actual generation last so that we don't accidentally wipe away the boss door.
    roomCount--; //exclude the boss room from further machinations

    //place room contents and ensure a minimum number of upgrades
    int numShinies = 0;
    vector<int> shinyRooms;
    int iters = 0;
    while (numShinies < 3) //flat number of upgrades for now... increase threshold when small keys are added
    {
        for(int n = 0; n < roomCount; n++)
        {
            int x = roomPositions[n].x;
            int y = roomPositions[n].y;
            if(rooms[x][y] != nullptr && rooms[x][y]->treasures == 0)
            {
                rooms[x][y]->removeEntities();
                rooms[x][y]->randomize();
                numShinies += rooms[x][y]->treasures;
                if (rooms[x][y]->treasures)
                {
                    shinyRooms.push_back(n);
                }
            }
        }
        iters++;
        if (iters > 10)
            printf("Oh god level generation is taking too long because Mark is a moron\n");
    }
    //...find a shiny and make it the boss key
    sf::Vector2i bossKeyPos = roomPositions[shinyRooms[rand()%shinyRooms.size()]];
    rooms[bossKeyPos.x][bossKeyPos.y]->placeBossKey();

    //if we're sufficiently deep, set the boss key room and all its children to be 'locked'
    //putting the door between the boss key room and its parent (or perhaps making sure at least 3 rooms are locked,
    //or something like that).  Then pick a shiny outside the locked area to turn into a key.

    //potentially repeat that if you're even deeper.

    rooms[bossPos.x][bossPos.y]->makeBossRoom(); //actually literally place the boss room.
}

Room* Level::getStartRoom()
{
    return startRoom;
}

Room* Level::getRoom(int x, int y)
{
    if(x < 0 || y < 0 || x >= w || y >= h)
    {
        return nullptr;
    }
    return rooms[x][y];
}

void Level::drawMinimap(sf::RenderTarget& window, float px, float py)
{
    sf::Sprite playerIcon(ResourceManager::getTexture("player_icon.png"));

    sf::RectangleShape roomBox({13, 13});
    roomBox.setFillColor(sf::Color::Black);
    roomBox.setOutlineColor(sf::Color::White);
    roomBox.setOutlineThickness(1.f);
    sf::RectangleShape linkBox({3, 3});
    linkBox.setFillColor(sf::Color::Black);
    for(int x = 0; x < w; x++)
    {
        for(int y = 0; y < h; y++)
        {
            sf::Vector2f drawPos(x * 16 + px,
                                 y * 16 + py);
            roomBox.setPosition(drawPos);
            if(!rooms[x][y] || !rooms[x][y]->explored)
            {
                roomBox.setOutlineColor(sf::Color(50, 50, 50));
                window.draw(roomBox);
                continue;
            }
            roomBox.setOutlineColor(sf::Color::White);
            window.draw(roomBox);

            if(rooms[x][y]->linkExists(Room::LINK_LEFT))
            {
                linkBox.setPosition(drawPos.x - 2, drawPos.y + 5);
                window.draw(linkBox);
            }
            if(rooms[x][y]->linkExists(Room::LINK_RIGHT))
            {
                linkBox.setPosition(drawPos.x + 12, drawPos.y + 5);
                window.draw(linkBox);
            }
            if(rooms[x][y]->linkExists(Room::LINK_UP))
            {
                linkBox.setPosition(drawPos.x + 5, drawPos.y - 2 );
                window.draw(linkBox);
            }
            if(rooms[x][y]->linkExists(Room::LINK_DOWN))
            {
                linkBox.setPosition(drawPos.x + 5, drawPos.y + 12);
                window.draw(linkBox);
            }

            if(rooms[x][y]->getPlayer())
            {
                playerIcon.setPosition(drawPos);
                window.draw(playerIcon);
            }
        }
    }
}

void Level::linkRooms(sf::Vector2i a, sf::Vector2i b)
{
    Room* roomA = rooms[a.x][a.y];
    Room* roomB = rooms[b.x][b.y];

    if(a.y == b.y)
    {
        // horizontal link
        if(a.x < b.x)
        {
            roomA->link(roomB, Room::LINK_RIGHT);
            roomB->link(roomA, Room::LINK_LEFT);
        }
        else
        {
            roomB->link(roomA, Room::LINK_RIGHT);
            roomA->link(roomB, Room::LINK_LEFT);
        }
    }
    else
    {
        // vertical link
        if(a.y < b.y)
        {
            roomA->link(roomB, Room::LINK_DOWN);
            roomB->link(roomA, Room::LINK_UP);
        }
        else
        {
            roomB->link(roomA, Room::LINK_DOWN);
            roomA->link(roomB, Room::LINK_UP);
        }
    }
}
