#include "ResourceManager.h"

const string ResourceManager::gfxPath = "gfx/";
const string ResourceManager::sfxPath = "sfx/";
const string ResourceManager::musicPath = "sfx/";
const string ResourceManager::fontPath = "gfx/font/";
const string ResourceManager::shaderPath = "gfx/shaders/";

map<string, sf::Texture*> ResourceManager::textures;
map<string, sf::Font*> ResourceManager::fonts;
map<string, sf::SoundBuffer*> ResourceManager::soundBuffers;
map<string, sf::Music*> ResourceManager::music;
map<string, sf::Shader*> ResourceManager::shaders;
list<sf::Sound*> ResourceManager::playingSounds;

vector<string> ResourceManager::songNames;

bool ResourceManager::muted = false;
float ResourceManager::musicVolume = 50;
float ResourceManager::sfxVolume = 50;

void ResourceManager::initialize()
{
    loadAllTextures();
    loadAllSounds();
    loadFont("Tiny_04b.ttf");
}

void ResourceManager::loadAllTextures()
{
    loadTexture("creep.png");
    loadTexture("player.png");
    loadTexture("player_icon.png");
    loadTexture("player_walk.png");
    loadTexture("player_attack.png");
    loadTexture("player_dead.png");
    loadTexture("player_sleep.png");
    loadTexture("player_fall.png");
    loadTexture("boss_door.png");
    loadTexture("stairs.png");
    loadTexture("projectiles.png");

    loadTexture("slash_effect.png");
    loadTexture("tileset.png");
    loadTexture("items.png");
    loadTexture("boss_cast.png");
    loadTexture("boss_dead.png");
    loadTexture("boss_idle.png");
    loadTexture("boss_walk.png");
    loadTexture("traps.png");
}

void ResourceManager::loadAllSounds()
{
    loadSound("attack.wav");
    loadSound("destroyed.wav");
    loadSound("floor_crumble.wav");
    loadSound("hit.wav");
    loadSound("magic_boom.wav");
    loadSound("magic_cast.wav");
    loadSound("magic_reflect.wav");
    loadSound("powerup.wav");
    loadSound("spike.wav");
    loadSound("arrow_shot.wav");
    //music
    loadMusic("bgm.ogg");
}

void ResourceManager::cleanUp()
{
    map<string, sf::Texture*>::iterator tex_it;
    for(tex_it = textures.begin(); tex_it != textures.end(); tex_it++)
    {
        delete tex_it->second;
        tex_it->second = NULL;
    }
    textures.clear();

    map<string, sf::Font*>::iterator font_it;
     for(font_it = fonts.begin(); font_it != fonts.end(); font_it++)
    {
        delete font_it->second;
        font_it->second = NULL;
    }
    fonts.clear();

    for(list<sf::Sound*>::iterator snd_it = playingSounds.begin(); snd_it != playingSounds.end(); snd_it++)
    {
        delete *snd_it;
    }
    playingSounds.clear();

    map<string, sf::SoundBuffer*>::iterator sfx_it;
    for(sfx_it = soundBuffers.begin(); sfx_it != soundBuffers.end(); sfx_it++)
    {
        delete sfx_it->second;
        sfx_it->second = NULL;
    }
    soundBuffers.clear();

    map<string, sf::Music*>::iterator mus_it;
    for(mus_it = music.begin(); mus_it != music.end(); mus_it++)
    {
        delete mus_it->second;
        mus_it->second = NULL;
    }
    music.clear();

    for_each(shaders.begin(), shaders.end(), [&](pair<string,sf::Shader*> p)
    {
        delete p.second;
        p.second = NULL;
    });
    shaders.clear();
}

bool ResourceManager::loadTexture(string filename)
{
    if(textures.find(filename) != textures.end())
    {
        printf("%s has already been loaded!\n", filename.c_str());
        return false;
    }
    sf::Texture* t = new sf::Texture();
    sf::err().rdbuf(NULL);
    if(!t->loadFromFile(gfxPath + filename))
    {
        if(!t->loadFromFile("../" + gfxPath + filename))
        {
            sf::err().rdbuf(std::cerr.rdbuf());
            printf("Error loading: %s!\n", filename.c_str());
            delete t; t = NULL;
            return false;
        }

    }
    sf::err().rdbuf(std::cerr.rdbuf());
    t->setSmooth(false);
    //printf("Loaded %s!\n", filename.c_str());
    textures.insert(pair<string, sf::Texture*>(filename, t));
    return true;
}

void ResourceManager::unloadTexture(string filename)
{
    map<string, sf::Texture*>::iterator tex_it = textures.find(filename);
    if(tex_it != textures.end())
    {
        delete tex_it->second;
        textures.erase(tex_it);
    }
}

bool ResourceManager::loadFont(string filename)
{
    if(fonts.find(filename) != fonts.end())
    {
        printf("%s has already been loaded!", filename.c_str());
        return false;
    }
    sf::Font* font = new sf::Font();
    sf::err().rdbuf(NULL);
    if(!font->loadFromFile(fontPath + filename))
    {
        if(!font->loadFromFile("../" + fontPath + filename))
        {
            sf::err().rdbuf(std::cerr.rdbuf());
            printf("Error loading: %s!\n", filename.c_str());
            delete font; font = NULL;
            return false;
        }
    }
    sf::err().rdbuf(std::cerr.rdbuf());
    const_cast<sf::Texture&>(font->getTexture(8)).setSmooth(false);
    const_cast<sf::Texture&>(font->getTexture(12)).setSmooth(false);
    const_cast<sf::Texture&>(font->getTexture(16)).setSmooth(false);
    fonts.insert(pair<string, sf::Font*>(filename, font));
    return true;
}

void ResourceManager::unloadFont(string filename)
{
    map<string, sf::Font*>::iterator font_it = fonts.find(filename);
    if(font_it != fonts.end())
    {
        delete font_it->second;
        fonts.erase(font_it);
    }
}

bool ResourceManager::loadSound(string filename)
{
    if(soundBuffers.find(filename) != soundBuffers.end())
    {
        printf("%s has already been loaded!", filename.c_str());
        return false;
    }
    sf::err().rdbuf(NULL);
    sf::SoundBuffer* soundBuff = new sf::SoundBuffer();
    if(!soundBuff->loadFromFile(sfxPath + filename))
    {
        if(!soundBuff->loadFromFile("../" + sfxPath + filename))
        {
            sf::err().rdbuf(std::cerr.rdbuf());
            printf("Error loading: %s!\n", filename.c_str());
            delete soundBuff;
            soundBuff = NULL;
            return false;
        }
    }
    sf::err().rdbuf(std::cerr.rdbuf());
    soundBuffers.insert(pair<string, sf::SoundBuffer*>(filename, soundBuff));
    return true;
}

void ResourceManager::unloadSound(string filename)
{
    map<string, sf::Font*>::iterator font_it = fonts.find(filename);
    if(font_it != fonts.end())
    {
        delete font_it->second;
        fonts.erase(font_it);
    }
}

bool ResourceManager::loadMusic(string filename)
{
    if(music.find(filename) != music.end())
    {
        printf("%s has already been loaded!", filename.c_str());
        return false;
    }

    sf::Music* m = new sf::Music();
    m->setLoop(true);
    m->setVolume(musicVolume);

    sf::err().rdbuf(NULL);
    if(!m->openFromFile(musicPath+filename))
    {
        if(!m->openFromFile("../" + musicPath + filename))
        {
            sf::err().rdbuf(std::cerr.rdbuf());
            printf("Error loading: %s!\n", filename.c_str());
            delete m; m = NULL;
            return false;
        }
    }
    sf::err().rdbuf(std::cerr.rdbuf());
    music.insert(pair<string, sf::Music*>(filename, m));
    return true;
}

bool ResourceManager::loadShader(string filename, sf::Shader::Type type)
{
    if(shaders.find(filename) != shaders.end())
    {
        printf("%s has already been loaded!", filename.c_str());
        return false;
    }
    sf::Shader* shader = new sf::Shader();
    sf::err().rdbuf(NULL);
    if(!shader->loadFromFile(shaderPath + filename, type))
    {
        if(!shader->loadFromFile("../" + shaderPath + filename, type))
        {
            sf::err().rdbuf(std::cerr.rdbuf());
            printf("Error loading: %s!\n", filename.c_str());
            delete shader; shader = NULL;
            return false;
        }
    }
    sf::err().rdbuf(std::cerr.rdbuf());
    shader->setParameter("texture", sf::Shader::CurrentTexture);
    shaders.insert(pair<string, sf::Shader*>(filename, shader));
    return true;
}

void ResourceManager::unloadMusic(string filename)
{
    map<string, sf::Music*>::iterator mus_it = music.find(filename);
    if(mus_it != music.end())
    {
        delete mus_it->second;
        music.erase(mus_it);
    }
}
void ResourceManager::unloadShader(string filename)
{
    map<string, sf::Shader*>::iterator shade_it = shaders.find(filename);
    if(shade_it != shaders.end())
    {
        delete shade_it->second;
        shaders.erase(shade_it);
    }
}

sf::Texture& ResourceManager::getTexture(string filename)
{
    if(textures.find(filename) == textures.end())
    {
        printf("ERROR: Unable to get texture: %s\n", filename.c_str()); fflush(stdout);
    }
    return *(textures.find(filename)->second);
}

sf::Font& ResourceManager::getDefaultFont()
{
    if(fonts.size() == 0)
        printf("ERROR: No default font found!\n");
    return *(fonts.begin()->second);
}

sf::Font& ResourceManager::getFont(string filename)
{
    return *(fonts.find(filename)->second);
}

sf::SoundBuffer& ResourceManager::getSoundBuffer(string filename)
{
    return *(soundBuffers.find(filename)->second);
}

void ResourceManager::playSound(string filename)
{
    if(muted)
        return;
    playingSounds.push_back(new sf::Sound(*(soundBuffers.find(filename)->second)));
    playingSounds.back()->play();
    playingSounds.back()->setVolume(sfxVolume);

    ///remove sounds
    list<sf::Sound*>::iterator snd_it = playingSounds.begin();
    for(;snd_it != playingSounds.end(); snd_it++)
    {
        if((*snd_it)->getStatus() == sf::Sound::Stopped)
        {
            delete *snd_it;
            playingSounds.erase(snd_it);
            snd_it = playingSounds.begin();
        }
    }
}

sf::Music& ResourceManager::getMusic(string filename)
{
    return *(music.find(filename)->second);
}

sf::Music* ResourceManager::getMusicPointer(string filename)
{
    return music.find(filename)->second;
}

void ResourceManager::playMusic(string filename)
{
    sf::Music* currMusic = getMusicPointer(filename);
    currMusic->play();
}

void ResourceManager::stopMusic(string filename)
{
    sf::Music* currMusic = getMusicPointer(filename);
    currMusic->stop();
}

void ResourceManager::stopAllMusic()
{
    map<string, sf::Music*>::iterator mus_it;
    for(mus_it = music.begin(); mus_it != music.end(); mus_it++)
    {
        mus_it->second->stop();
    }
}

void ResourceManager::toggleMute()
{
    muted = !muted;
    if(muted)
    {
        map<string, sf::Music*>::iterator mus_it;
        for(mus_it = music.begin(); mus_it != music.end(); mus_it++)
        {
            mus_it->second->setVolume(0);
        }
    }
    else
    {
        map<string, sf::Music*>::iterator mus_it;
        for(mus_it = music.begin(); mus_it != music.end(); mus_it++)
        {
            mus_it->second->setVolume(musicVolume);
        }
    }
}

void ResourceManager::setSfxVolume(float volume)
{
    sfxVolume = volume;
}

void ResourceManager::setMusicVolume(float volume)
{
    if(volume != musicVolume)
    {
        musicVolume = volume;

        map<string, sf::Music*>::iterator mus_it;
        for(mus_it = music.begin(); mus_it != music.end(); mus_it++)
        {
            mus_it->second->setVolume(musicVolume);
        }
    }
}

bool ResourceManager::isMuted()
{
    return muted;
}

sf::Shader* ResourceManager::getShaderPointer(string filename)
{
    assert(shaders.find(filename) != shaders.end());
    return shaders.find(filename)->second;
}
