#include "Room.h"
#include "RoomDatabase.h"
#include "Entity.h"
#include "Player.h"
#include "Creep.h"
#include "Boss.h"
//#define DEBUG_ITEMS

const int Room::WIDTH = 18;
const int Room::HEIGHT = 18;
const int Room::TILE_SIZE = 16;

const int Room::SPAWN_RATES[12] = {40,45,50,60,70,75,80,85,89,93,96,99};

Room::Room(int x, int y, int depth, Room* parent):
    x(x),
    y(y),
    parent(parent),
    depth(depth),
    w(WIDTH),
    h(HEIGHT),
    explored(false),
    treasures(0),
    hasBoss(false),
    tilesetSprite(ResourceManager::getTexture("tileset.png"), 16, 16, -1, 0, 0, 0, false),
    player(nullptr)

{
    tilesetSprite.setOrigin(0, 0);
    links.resize(4, NULL);

    tiles = new Tile*[w];
    for(int x = 0; x < w; x++)
    {
        tiles[x] = new Tile[h];
    }
    particles.resize(100, nullptr);
}

Room::~Room()
{
    for(int x = 0; x < w; x++)
    {
        delete[] tiles[x];
    }
    delete[] tiles;
}

void Room::addEntity(Entity* other)
{
    other->setRoom(this);
    entities.push_back(other);
    if(other->entityType == Entity::PLAYER)
    {
        player = static_cast<Player*>(other);
    }
}

void Room::addParticle(Particle* p)
{
    bool added = false;
    for(unsigned i = 0; i < particles.size(); i++)
    {
        if(particles[i] == nullptr)
        {
            particles[i] = p;
            added = true;
            break;
        }
    }

    if(!added)
    {
        delete p;
    }
}

void Room::removeEntity(Entity* other)
{
    for(unsigned i = 0; i < entities.size(); i++)
    {
        if(entities[i] == other)
        {
            if(other->entityType == Entity::PLAYER)
            {
                player = nullptr;
            }
            entities.erase(entities.begin() + i);
            break;
        }
    }
}

void Room::removeEntities() //I'm so sorry -Mark
{
    for(unsigned i = entities.size(); i > 0; i--)
    {
        Entity* e = entities[i-1];
        if(e->entityType == Entity::PLAYER)
        {
            player = nullptr;
        }
        else
        {
            delete e;
        }
        entities[i-1] = NULL;
        entities.erase(entities.begin() + i - 1);
    }
}

Player* Room::getPlayer()
{
    return player;
}

void Room::step()
{
    for(unsigned i = 0; i < entities.size(); i++)
    {
        entities[i]->step();
    }

    for(unsigned i = 0; i < entities.size(); i++)
    {
        Entity* e = entities[i];
        if(e->die && e->removeOnDie)
        {
            delete e;
            entities[i] = NULL;
            entities.erase(entities.begin() + i);
            i-=1;
            continue;
        }
    }

    for(unsigned i = 0; i < entities.size(); i++)
    {
        for(unsigned j = i+1; j < entities.size(); j++)
        {
            entities[i]->doCollide(entities[j]);
            entities[j]->doCollide(entities[i]);
        }
    }
    if(hasBoss)
    {
        ///TODO: Change this to the boss and make the boss spawn the stairs (and some hp?) on death
        if(boss->die && !stairsSpawned)
        {
            float stairX = (WIDTH/2-1)*TILE_SIZE;
            float stairY = (HEIGHT/2-1)*TILE_SIZE;
            addEntity(new Stairs(stairX,stairY));
            stairsSpawned = true;
        }
    }
}

void Room::draw(sf::RenderTarget& window, bool debug)
{

    sf::RectangleShape flagBox({2, 2});
    flagBox.setFillColor(sf::Color::Green);
    for(int x = 0; x < w; x++)
    {
        for(int y = 0; y < h; y++)
        {
            if(tiles[x][y].displayIndex >= 0)
            {
                tilesetSprite.setPosition(x * TILE_SIZE, y * TILE_SIZE);
                tilesetSprite.setFrame(tiles[x][y].displayIndex);
                tilesetSprite.draw(window);
            }
        }
    }

    //slow but no need for Z-index
    for(unsigned i = 0; i < entities.size(); i++)
    {
        if(entities[i]->entityType != Entity::TRAP) continue;
        entities[i]->draw(window);
        if(debug)
        {
            entities[i]->drawBoundingBox(window);
        }
    }
    for(unsigned i = 0; i < entities.size(); i++)
    {
        if(entities[i]->entityType == Entity::TRAP) continue;
        entities[i]->draw(window);
        if(debug)
        {
            entities[i]->drawBoundingBox(window);
        }
    }


    for(unsigned i = 0; i < particles.size(); i++)
    {
        if(particles[i])
        {
            particles[i]->draw(window);
            if(particles[i]->die)
            {
                delete particles[i];
                particles[i] = nullptr;
            }
        }
    }
}

bool Room::isSolid(int x, int y)
{
    return tiles[x][y].solid;
}

Tile** Room::getTiles()
{
    return tiles;
}
Tile& Room::getTile(int x, int y)
{
    return tiles[x][y];
}

void Room::randomize()
{
    ///debug construct walls
    for(int x = 0; x < w; x++)
    {
        for(int y = 0; y < h; y++)
        {
            if(x == 0 || y == 0 || x == w-1 || y == h-1)// ||
               //rand()%12 == 0)
            {
                tiles[x][y].type = TILE_SOLID;
            }
            #ifdef DEBUG_ITEMS
            if (!hasBoss && x >= 3 && x <= 6)
            {
                if (y > 2 && y < h-3)
                    tiles[x][y].type = TILE_REWARD;
            }
            #endif
        }
    }

    int linkSum = 0;
    if(links[LINK_LEFT])
    {
        for(int y = 8; y <= 9; y++)
        {
            tiles[0][y].type = TILE_NONE;
        }
        linkSum += 1;
    }
    if(links[LINK_UP])
    {
        for(int x = 8; x <= 9; x++)
        {
            tiles[x][0].type = TILE_NONE;
        }
        linkSum += 2;
    }
    if(links[LINK_RIGHT])
    {
        for(int y = 8; y <= 9; y++)
        {
            tiles[w-1][y].type = TILE_NONE;
        }
        linkSum += 4;
    }
    if(links[LINK_DOWN])
    {
        for(int x = 8; x <= 9; x++)
        {
            tiles[x][h-1].type = TILE_NONE;
        }
        linkSum += 8;
    }

    if (parent && !hasBoss) //not a starting room or a boss room
    {
        RoomDatabase::getRandomRoomLayout(this, linkSum);
    }
    else if (hasBoss)
    {
        boss = new Boss((WIDTH/2-1)*TILE_SIZE,
                        (HEIGHT/2-1)*TILE_SIZE, depth);
        addEntity(boss);
        stairsSpawned = false;
    }
    generateCollisionFlags();
    generateDisplayIndices();
    generateTileEntities();
    #ifdef DEBUG_ITEMS
    if (!parent)
    {
        placeBossKey();
    }
    #endif
}


void Room::generateCollisionFlags()
{
    //First set bool flags
    for (int x = 0; x < w; x++)
    {
        for (int y = 0; y < h; y++)
        {
            tiles[x][y].solid = false;
            tiles[x][y].left = false;
            tiles[x][y].up = false;
            tiles[x][y].right = false;
            tiles[x][y].down = false;
            int type = tiles[x][y].type;
            switch(type)
            {
                case TILE_SOLID:
                    tiles[x][y].solid = true;
                    break;
                default:
                    break;
            }
        }
    }
    // After basic flags, set directional flags
    for (int x = 0; x < w; x++)
    {
        for (int y = 0; y < h; y++)
        {
            int type = tiles[x][y].type;
            switch(type)
            {
                case TILE_SOLID:
                    if(x > 0 && !isSolid(x-1, y))
                    {
                        tiles[x][y].left = true;
                    }
                    if(x < w-1 && !isSolid(x+1, y))
                    {
                        tiles[x][y].right = true;
                    }
                    if(y > 0 && !isSolid(x, y-1))
                    {
                        tiles[x][y].up = true;
                    }
                    if(y < h-1 && !isSolid(x, y+1))
                    {
                        tiles[x][y].down = true;
                    }
                    break;
            }
        }
    }
}

void Room::generateDisplayIndices()
{
    for (int x = 0; x < w; x++)
    {
        for (int y = 0; y < h; y++)
        {
            tiles[x][y].displayIndex = -1;
            switch(tiles[x][y].type)
            {
                case TILE_SOLID:
                    tiles[x][y].displayIndex = 0;
                    if(x > 0 && tiles[x-1][y].type != TILE_SOLID)
                    {
                        tiles[x][y].displayIndex += 1;
                    }
                    if(x < w-1 && tiles[x+1][y].type != TILE_SOLID)
                    {
                        tiles[x][y].displayIndex += 4;
                    }
                    if(y > 0 && tiles[x][y-1].type != TILE_SOLID)
                    {
                        tiles[x][y].displayIndex += 2;
                    }
                    if(y < h-1 && tiles[x][y+1].type != TILE_SOLID)
                    {
                        tiles[x][y].displayIndex += 8;
                    }
                    break;
            }
        }
    }
}

void Room::generateTileEntities()
{
    int spawnCounter = rand()%100;

    for (int x = 0; x < w; x++)
    {
        for (int y = 0; y < h; y++)
        {
            switch(tiles[x][y].type)
            {
            case TILE_MONSTER_SPAWN:
                if (depth >= 12)
                {
                    addEntity(new Creep(x * TILE_SIZE, y * TILE_SIZE));
                }
                else
                {
                    spawnCounter += SPAWN_RATES[depth];
                    if (spawnCounter > 100)
                    {
                        spawnCounter -= 100;
                        addEntity(new Creep(x * TILE_SIZE, y * TILE_SIZE));
                    }
                }
                break;
            case TILE_REWARD:
                {
                Item::ItemType type = Item::ITEM_HEALTH;
                if (rand()%10 < 6)
                {
                    if (rand()%10 < 5)
                    {
                        type = Item::ITEM_ATTACK_SPEED;
                    }
                    else
                    {
                        type = Item::ITEM_RUN_SPEED;
                    }
                }
                addEntity(new Item(x*TILE_SIZE,y*TILE_SIZE, type));
                } //bizarre scope block is so we don't cross initializations of type
                treasures++;
                break;
            case TILE_ARROWTRAP:
                addEntity(new Trap(Trap::ARROWTRAP,x*TILE_SIZE,y*TILE_SIZE));
                break;
            case TILE_SPIKE:
                addEntity(new Trap(Trap::SPIKETRAP,x*TILE_SIZE,y*TILE_SIZE));
                break;
            case TILE_HOLE:
                addEntity(new Trap(Trap::HOLETRAP,x*TILE_SIZE,y*TILE_SIZE));
                break;
            }
        }
    }
}

void Room::placeBossKey()
{
    float x = -1.0;
    float y = -1.0;
    for (unsigned i = 0; i < entities.size(); i++)
    {
        Entity* e = entities[i];
        if (e->entityType == Entity::ITEM)
        {
            x = e->x;
            y = e->y;
            delete e;
            entities.erase(entities.begin() + i);
            break;
        }
    }
    assert(x > -1);
    addEntity(new Item(x,y,Item::ITEM_BOSS_KEY));
}

void Room::makeBossRoom()
{
    hasBoss = true;
    int px = parent->x;
    int py = parent->y;
    float doorX,doorY;
    bool vertical; //true if long side of door is vertical
    if (py == y)
    {
        vertical = true; //link is horizontal.
        doorY = Room::HEIGHT/2 - 1;
        if (px < x) //parent is left of boss room
        {
            doorX = Room::WIDTH - 1;
        }
        else
        {
            doorX = 0.0;
        }
    }
    else
    {
        vertical = false;
        doorX = Room::WIDTH/2 - 1;
        if (py < y) //parent is above boss room
        {
            doorY = Room::HEIGHT - 1;
        }
        else
        {
            doorY = 0.0;
        }
    }
    parent->addEntity(new BossDoor(doorX*TILE_SIZE,doorY*TILE_SIZE,vertical));
    randomize();
}
bool Room::findCreeps()
{
    for(int c = 0; c < entities.size(); c++)
    {
        if(entities[c]->entityType == Entity::CREEP && !entities[c]->die)
        {
            return true;
        }
    }
    return false;
}

void Room::link(Room* room, LinkDirection dir)
{
    assert(!links[dir]);
    links[dir] = room;

    switch(dir)
    {
    case LINK_LEFT:
        break;
    case LINK_RIGHT:
        break;
    case LINK_UP:
        break;
    case LINK_DOWN:
        break;
    }
}

bool Room::linkExists(LinkDirection dir)
{
    return links[dir] != NULL;
}
